# -*- coding: utf-8 -*-
"""
The shared library can also be built manually using the command:
$ cythonize -X language_level=3 -a -i ./iscc/cdc.py
$ cythonize -X language_level=3 -a -i ./iscc/minhash.py
"""


def build(setup_kwargs):
    try:
        from Cython.Build import cythonize, build_ext

        setup_kwargs.update(
            dict(
                ext_modules=cythonize(["./iscc/cdc.py", "./iscc/minhash.py"]),
                cmdclass=dict(build_ext=build_ext),
            )
        )
    except Exception as e:
        print(e)
        print("************************************************************")
        print("Cannot compile C accelerator module, use pure python version")
        print("************************************************************")



import cython
from libc.stdint cimport uint64_t, uint32_t, uint8_t


cdef uint64_t MAXI64
cdef uint64_t MPRIME
cdef uint32_t MAXH
cdef uint64_t[64] MPA
cdef uint64_t[64] MPB

@cython.locals(a=uint64_t, b=uint64_t, f=uint64_t)
cpdef list minhash(list features)

cpdef bytes minhash_64(list features)

cpdef bytes minhash_256(list features)

@cython.locals(bits=str, bitpos=uint8_t, h=uint64_t)
cpdef bytes compress(list mhash, int lsb=*)



# -*- coding: utf-8 -*-
"""
  ffmpeg signature parsing based on binary_export function from:
     https://www.ffmpeg.org/doxygen/3.4/vf__signature_8c_source.html

"""
from dataclasses import dataclass
from fractions import Fraction
from typing import Tuple, List
import numpy as np
from loguru import logger

# try loading optional jit compiler for speedup
try:
    from numba import jit, njit

    if __name__ == "__main__":
        logger.debug("using numba engine for jit compiling")
except ImportError:

    def jit(func):
        def do_nothing(*args, **kwargs):
            return func(*args, **kwargs)

        return do_nothing

    if __name__ == "__main__":
        logger.debug("can not load numba lib")


@dataclass
class Frame:
    """datatype to return by read_ffmpeg_signature"""

    vector: np.ndarray  # 380 dimensional vector, range: 0..2
    elapsed: Fraction  # time elapsed since start of video
    confidence: int  # signature confidence, range: 0..255


@jit
def to_binary(byte_data: bytes) -> np.ndarray:
    """convert each byte into a 8 bytes containing 1/0 value"""
    np_byte_data = np.frombuffer(byte_data, dtype=np.uint8)
    binary_data = np.zeros((len(byte_data) * 8,), dtype=np.uint8)
    p = 0
    for i, b in enumerate(np_byte_data):
        for ii in range(7, -1, -1):
            binary_data[p + ii] = b & 1
            b >>= 1
        p += 8
    return binary_data


@jit
def pop_bits(data_bits: np.ndarray, pos: int, bits: int = 32) -> Tuple[int, int]:
    """take out 0/1 values and pack them again to an unsigned integer
    :param data_bits: 0/1 data
    :param pos: position in 0/1 data
    :param bits: number of bits
    :return: value, new position
    """
    value = 0
    for i in range(bits):
        value = value + value + data_bits[pos + i]
    pos += bits
    return value, pos


@jit
def pop_bit(data_bits: np.ndarray, pos: int) -> Tuple[int, int]:
    """take out 1 bit of 0/1 values
    :param data_bits: 0/1 data
    :param pos: position in 0/1 data
    :return: take out just one bit
    """
    value = data_bits[pos]
    pos += 1
    return value, pos


@jit
def calc_byte_to_bit3() -> np.ndarray:
    """lookup table
    :return: table to convert a 8bit value into five three-bit-values
    """
    table_3_bit = np.zeros((256, 5), dtype=np.uint8)
    for i in range(256):
        div3 = 3 * 3 * 3 * 3
        for iii in range(0, 5):
            table_3_bit[i, iii] = (i // div3) % 3
            div3 //= 3
    return table_3_bit


SIGELEM_SIZE = 380


@jit
def nada(_):
    pass


def _read_ffmpeg_signature(
    byte_data: bytes, test_mode: bool, debug=nada
) -> Tuple[List[np.ndarray], List[int], List[int], List[int]]:
    """read data from the binary FFMpeg signature
        return lists of: vectors, elapsed time, confidence
        there is one entry for each frame.
    :param byte_data: actual ffmpeg signature data
    :param test_mode: basic data assert verification, print out details
    :return: [vector],[elapsed time],[confidence]
    """

    table_3_bit = calc_byte_to_bit3()
    data_bits = to_binary(byte_data)
    pos = 0

    if test_mode:
        # put_bits32(&buf, 1); /* NumOfSpatial Regions, only 1 supported */
        _, pos = pop_bits(data_bits, pos)
        assert _ == 1
        # put_bits(&buf, 1, 1); /* SpatialLocationFlag, always the whole image */
        debug("expect 1 " + str(pop_bits(data_bits, pos, 1)[0]))
        pos += 1
        # put_bits32(&buf, 0); /* PixelX,1 PixelY,1, 0,0 */
        debug("expect 0 " + str(pop_bits(data_bits, pos)[0]))
        pos += 32
        # put_bits(&buf, 16, sc->w-1 & 0xFFFF); /* PixelX,2 */
        debug("expect X " + str(pop_bits(data_bits, pos, 16)[0]))
        pos += 16
        # put_bits(&buf, 16, sc->h-1 & 0xFFFF); /* PixelY,2 */
        debug("expect Y " + str(pop_bits(data_bits, pos, 16)[0]))
        pos += 16
        # put_bits32(&buf, 0); /* StartFrameOfSpatialRegion */
        debug("expect StartFrameOfSpatialRegion " + str(pop_bits(data_bits, pos)[0]))
        pos += 32
        # put_bits32(&buf, sc->lastindex); /* NumOfFrames */
    else:
        pos += 129

    num_of_frames, pos = pop_bits(data_bits, pos)
    media_time_unit, pos = pop_bits(data_bits, pos, 16)
    if test_mode:
        debug("expect MediaTimeUnit on second takes x units: " + str(media_time_unit))
        debug("expect num_of_frames " + str(num_of_frames))
        # /* hoping num is 1, other values are vague */
        # /* den/num might be greater than 16 bit, so cutting it */
        # put_bits(&buf, 16, 0xFFFF & (sc->time_base.den / sc->time_base.num)); /* MediaTimeUnit */
    if test_mode:
        # put_bits(&buf, 1, 1); /* MediaTimeFlagOfSpatialRegion */
        debug(
            "expect MediaTimeFlagOfSpatialRegion " + str(pop_bits(data_bits, pos, 1)[0])
        )
        pos += 1
        # put_bits32(&buf, 0); /* StartMediaTimeOfSpatialRegion */
        _, pos = pop_bits(data_bits, pos)
        assert _ == 0
        # put_bits32(&buf, 0xFFFFFFFF & sc->coarseend->last->pts); /* EndMediaTimeOfSpatialRegion */
        debug("expect  " + str(pop_bits(data_bits, pos)[0]))
        pos += 32
        # put_bits32(&buf, numofsegments); /* NumOfSegments */
    else:
        pos += 1 + 32 + 32

    num_of_segments, pos = pop_bits(data_bits, pos)
    if test_mode:
        debug("expect num_of_segments " + str(num_of_segments))
    pos += num_of_segments * (4 * 32 + 1 + 5 * 243)

    if test_mode:
        # /* fine signatures */
        # put_bits(&buf, 1, 0); /* CompressionFlag, only 0 supported */
        debug("expect compression " + str(pop_bits(data_bits, pos, 1)[0]))
        pos += 1
    else:
        pos += 1

    frame_sigs_v = []
    frame_sigs_c = []
    frame_sigs_e = []
    frame_sigs_tu = []
    # for (fs = sc->finesiglist; fs; fs = fs->next) {
    for i in range(num_of_frames):
        if test_mode:
            # put_bits(&buf, 1, 1); /* MediaTimeFlagOfFrame */
            _, pos = pop_bits(data_bits, pos, 1)
            assert _ == 1
            # put_bits32(&buf, 0xFFFFFFFF & fs->pts); /* MediaTimeOfFrame */
            raw_media_time, pos = pop_bits(data_bits, pos)

            debug("MediaTimeOfFrame  " + str(raw_media_time // media_time_unit))
            # put_bits(&buf, 8, fs->confidence); /* FrameConfidence */
            frame_confidence, pos = pop_bits(data_bits, pos, 8)
            debug("FrameConfidence " + str(frame_confidence))
        else:
            pos += 1
            raw_media_time, pos = pop_bits(data_bits, pos)
            frame_confidence, pos = pop_bits(data_bits, pos, 8)

        if test_mode:
            s = ""
            # for (i = 0; i < 5; i++)
            for ii in range(5):
                # put_bits(&buf, 8, fs->words[i]); /* Words */
                s += str(pop_bits(data_bits, pos, 8)[0]) + " "
                pos += 8
            debug(s)
        else:
            pos += 5 * 8

        vec = np.zeros((SIGELEM_SIZE,), dtype=np.uint8)
        p = 0
        for ii in range(SIGELEM_SIZE // 5):
            dat, pos = pop_bits(data_bits, pos, 8)
            vec[p : p + 5] = table_3_bit[dat]
            p += 5
        frame_sigs_v.append(vec)
        frame_sigs_e.append(raw_media_time)
        frame_sigs_c.append(frame_confidence)
        frame_sigs_tu.append(media_time_unit)
        if test_mode:
            debug(vec)
            # /* frame signature */
            # for (i = 0; i < SIGELEM_SIZE/5; i++) {
            #     put_bits(&buf, 8, fs->framesig[i]);

    return frame_sigs_v, frame_sigs_e, frame_sigs_c, frame_sigs_tu


@jit
def _read_ffmpeg_signature_jit(
    byte_data: bytes,
) -> Tuple[List[np.ndarray], List[int], List[int], List[int]]:
    """read data from the binary FFMpeg signature
        return lists of: vectors, elapsed time, confidence
        there is one entry for each frame.
    :param byte_data: actual ffmpeg signature data
    :param test_mode: basic data assert verification, print out details
    :return: [vectors],[elapsed time],[confidence]
    """

    table_3_bit = calc_byte_to_bit3()
    data_bits = to_binary(byte_data)
    pos = 0
    pos += 129
    num_of_frames, pos = pop_bits(data_bits, pos)
    media_time_unit, pos = pop_bits(data_bits, pos, 16)
    pos += 1 + 32 + 32
    num_of_segments, pos = pop_bits(data_bits, pos)
    pos += num_of_segments * (4 * 32 + 1 + 5 * 243)
    pos += 1
    frame_sigs_v = []
    frame_sigs_c = []
    frame_sigs_e = []
    frame_sigs_tu = []
    # for (fs = sc->finesiglist; fs; fs = fs->next) {
    for i in range(num_of_frames):
        pos += 1
        raw_media_time, pos = pop_bits(data_bits, pos)
        frame_confidence, pos = pop_bits(data_bits, pos, 8)
        pos += 5 * 8
        vec = np.zeros((SIGELEM_SIZE,), dtype=np.uint8)
        p = 0
        for ii in range(SIGELEM_SIZE // 5):
            dat, pos = pop_bits(data_bits, pos, 8)
            vec[p : p + 5] = table_3_bit[dat]
            p += 5
        frame_sigs_v.append(vec)
        frame_sigs_e.append(raw_media_time)
        frame_sigs_c.append(frame_confidence)
        frame_sigs_tu.append(media_time_unit)

    return frame_sigs_v, frame_sigs_e, frame_sigs_c, frame_sigs_tu


def read_ffmpeg_signature(byte_data: bytes, test_mode=False) -> List[Frame]:
    """wrapper for _read_ffmpeg_signature as numba code can not return a dataclass

    :param byte_data: actual ffmpeg signature data
    :param test_mode: basic data assert verification, print out details
    :return: [FFMpegFrameSignature()]
    """
    if test_mode:
        l = _read_ffmpeg_signature(byte_data, test_mode, debug=logger.debug)
    else:
        l = _read_ffmpeg_signature_jit(byte_data)
    frame_signatures = []
    for v, e, c, tu in zip(*l):
        frame_signatures.append(Frame(vector=v, elapsed=Fraction(e, tu), confidence=c))
    return frame_signatures


if __name__ == "__main__":
    signature_byte_data = open("../tests/ffmpeg_signature.bin", "rb").read()

    logger.debug("run speed test")
    import time

    # without the jit compiling..
    # _ = read_ffmpeg_signature(byte_data=signature_byte_data)

    start = time.time()
    frame_signature = read_ffmpeg_signature(byte_data=signature_byte_data)
    duration = time.time() - start
    logger.debug(
        f"extracted signatures from {len(frame_signature)} frames in {duration:.3f} seconds."
    )
    logger.debug(f"{1e9 * duration / len(frame_signature):.0f} nano seconds/frame.")



# -*- coding: utf-8 -*-
from loguru import logger
import unicodedata
from os.path import basename, splitext
from typing import Any, Generator, Union
from urllib.parse import urlparse
import langdetect
import xxhash
from functools import lru_cache
import langcodes
from iscc.schema import FeatureType, Options
from iscc.cdc import data_chunks
from iscc import uread
from iscc.utils import sliding_window
from iscc.codec import encode_base64
from iscc.schema import Readable
from iscc.mediatype import mime_clean, mime_to_gmt
from iscc.minhash import minhash_64, minhash_256
from tika import parser as tika_parser


# Set for deterministic language detection
langdetect.DetectorFactory.seed = 0


# Common Control Characters considered whitespace
CC_WHITESPACE = (
    "\u0009",  # Horizontal Tab (TAB)
    "\u000A",  # Linefeed (LF)
    "\u000D",  # Carriage Return (CR)
)


# Unicode categories to remove during text normalization
UNICODE_FILTER = frozenset(
    {
        "Cc",
        "Cf",
        "Cn",
        "Co",
        "Cs",
        "Mc",
        "Me",
        "Mn",
        "Pc",
        "Pd",
        "Pe",
        "Pf",
        "Pi",
        "Ps",
    }
)


def extract_text(data):
    # type: (Readable) -> str
    """Extract plaintext from a text document file."""
    text = _extract_with_tika(data).get("content", "")
    return text or ""


def extract_text_metadata(data, **options):
    # type: (Readable, **Any) -> dict
    """Extract metadata from text document (title, language, characters).

    :param data: File with textual content
    :key text_guess_title: Guess title from content if not found in metadata.
    """
    opts = Options(**options)
    file = uread.open_data(data)
    tika_result = _extract_with_tika(file)

    result = {}

    # Aquire title
    title = _title_from_tika(tika_result, **opts.dict())
    if title:
        result["title"] = title

    txt_raw = tika_result.get("content") or ""
    txt_norm = normalize_text(txt_raw)

    # Number of characters
    result["characters"] = len(txt_norm)

    if not txt_norm:
        return result

    try:
        lang = langdetect.detect(txt_norm)
        logger.debug(f"Detected langauge: {lang}")
        result["language"] = langcodes.standardize_tag(lang)
    except Exception as e:
        logger.warning(f"Language detection failed: {e}")

    return result


def extract_text_features(text, **options):
    # type: (str, **Any) -> dict
    """
    Create granular fingerprint for text (minhashes over ngrams from cdc-chunks).
    Text should be normalized before extracting text features.

    :param str text: Normalized plaintext.
    :key text_avg_chunk_size: Avg. number of chars per text chunk to be hashed.
    :key text_ngram_size: Sliding window size in number of characters.
    :returns dict: Dictionary with 'sizes' and 'features'.
    """
    opts = Options(**options)
    text = text.lower()
    chunks = chunk_text(text, text_avg_chunk_size=opts.text_avg_chunk_size)
    sizes = []
    feats = []
    for chunk in chunks:
        ngrams = (
            "".join(chars) for chars in sliding_window(chunk, opts.text_ngram_size)
        )
        features = [xxhash.xxh32_intdigest(s.encode("utf-8")) for s in ngrams]
        minimum_hash_digest = minhash_64(features)
        sizes.append(len(chunk))
        feats.append(encode_base64(minimum_hash_digest))
    return dict(kind=FeatureType.text, features=feats, sizes=sizes)


def normalize_text(text):
    # type: (Union[str, bytes], bool) -> str
    """Unicode normalization and character filtering."""

    # 1. Convert bytes to str
    if isinstance(text, bytes):
        text = text.decode("utf-8")

    # 2. Remove leading/trailing whitespace
    text_stripped = text.strip()

    # 3. Decompose with NFD
    text_decomposed = unicodedata.normalize("NFD", text_stripped)

    # 4. Filter
    chars = []
    for c in text_decomposed:
        cat = unicodedata.category(c)
        if cat not in UNICODE_FILTER:
            chars.append(c)
        elif c in CC_WHITESPACE:
            chars.append(c)
    text_filtered = "".join(chars)

    # 5. Collapse consecutive whitespace
    wsproc_text = " ".join(text_filtered.split())

    # 6. Recombine
    recombined = unicodedata.normalize("NFKC", wsproc_text)

    return recombined


def hash_text(text, **options):
    # type: (str, **Any) -> bytes
    """
    Create a 256-bit similarity preserving hash for text input.
    Text should be normalized before hash creation.
    """
    opts = Options(**options)
    text = text.lower()
    ngrams = ("".join(chars) for chars in sliding_window(text, opts.text_ngram_size))
    features = [xxhash.xxh32_intdigest(s.encode("utf-8")) for s in ngrams]
    shash = minhash_256(features)
    return shash


def chunk_text(text, **options):
    # type: (str, **Any) -> Generator[str]
    """
    Generates variable sized text chunks (without leading BOM)

    :param text: normalized plaintext
    :key: text_avg_chunk_size: Targeted average size of text chunks in bytes.
    """
    opts = Options(**options)
    avg_size = opts.text_avg_chunk_size
    data = text.encode("utf-32-be")
    avg_size *= 4  # 4 bytes per character
    for chunk in data_chunks(data, utf32=True, text_avg_chunk_size=avg_size):
        yield chunk.decode("utf-32-be")


def trim_text(text, nbytes):
    # type: (str, int) -> str
    """Trim text such that its utf-8 encoded size does not exceed nbytes."""
    return text.encode("utf-8")[:nbytes].decode("utf-8", "ignore").strip()


@lru_cache(typed=True)
def _extract_with_tika(data):
    # type: (Readable) -> dict
    """Extract text and metadata from a 'text'-file via Tika.
    Result:
        {"content": "...", "metadata": "..."}
    """
    file = uread.open_data(data)
    buffer = file.read()
    return tika_parser.from_buffer(buffer)


def _title_from_tika(tika_result, **options):
    # type: (dict, **Any) -> str
    """Extract title from tika result.

    Extraction is atempted in the following order
        - try to find title in tika metadata
        - use first line from text content

    :param tika_result: result from tika parsing
    :key: text_guess_title: whether to guess the title from the text itself as fallback.
    :key: meta_trim_title: Max number of bytes for utf-8 encoded title.
    """
    opts = Options(**options)
    title = ""
    meta = tika_result.get("metadata")
    mime_type = mime_clean(meta.get("Content-Type"))
    gmt = mime_to_gmt(mime_type)

    if meta:
        title = meta.get("dc:title", "")
        title = title[0].strip() if isinstance(title, list) else title.strip()
        if not title:
            title = meta.get("title", "")
            title = title[0].strip() if isinstance(title, list) else title.strip()

    # See if string would survive normalization
    norm_title = normalize_text(title)

    # Falback to get title from content
    if not norm_title and opts.text_guess_title and gmt == "text":
        content = tika_result.get("content", "")
        if content is not None:
            first_line = content.strip().splitlines()[0]
            title = trim_text(normalize_text(first_line), opts.meta_trim_title)

    return title


def name_from_uri(uri):
    """Extract "filename" part of an uri without file extension to be uses as fallback
    title for an asset if no title information can be aquired.
    """
    result = urlparse(uri)
    base = basename(result.path)
    name = splitext(base)[0]
    name = name.replace("-", " ")
    name = name.replace("_", " ")
    return name



# -*- coding: utf-8 -*-
"""ISCC Reference Implementation"""
import json
from json import JSONDecodeError
from os.path import basename
import base64
import io
from PIL.ImageOps import exif_transpose
from humanize import naturalsize
from loguru import logger
from typing import List, Optional, Union, Any
from PIL import Image
import xxhash
from blake3 import blake3
from iscc.minhash import minhash_256
from pyld import jsonld
from iscc import jcs
from iscc import text, image, audio, video
from iscc.codec import (
    Code,
    MT,
    ST,
    ST_CC,
    VS,
    encode_base32,
    write_header,
    compose,
)
from iscc.cdc import data_chunks
from iscc.mp7 import read_ffmpeg_signature
from iscc.meta import meta_hash
from iscc.schema import (
    GMT,
    Options,
    Uri,
    Data,
    File,
    Readable,
)
from iscc.mediatype import mime_guess, mime_to_gmt
from iscc import uread
from iscc.data import extract_data_features, encode_data_features

from iscc.jldloader import requests_document_loader

jsonld.set_document_loader(requests_document_loader())

###############################################################################
# High-Level ISCC Code generator functions                                   #
###############################################################################


def code_iscc(uri, title=None, extra=None, **options):
    # type: (Union[Uri, File], Optional[str, dict], Optional[exec()], **Any) -> dict
    """Create a full ISCC.

    The full ISCC is a composite of Meta, Content, Data and Instance Codes.

    :param uri: File or filepath used for ISCC creation.
    :param title: Title of media asset (defaults to extracted metadata or filename)
    :param extra: Metadata to be used for Meta-Code generation
    :param options: See iscc.schema.Options for detailed ISCC generator options.
    """

    result = {"version": "0-0-0"}
    features = []

    file_obj = uread.open_data(uri)
    try:
        file_name = getattr(file_obj, "name", None)
        if file_name:
            result["filename"] = basename(file_name)

        instance = code_instance(file_obj, **options)
        result.update(instance)

        data = code_data(file_obj, **options)
        if "features" in data:
            features.append(data.pop("features"))
        result.update(data)

        content = code_content(file_obj, **options)
        if "features" in content:
            features.append(content.pop("features"))
        result.update(content)

        if features:
            result["features"] = features

        if not title:
            title = content.get("title")
        if not title and file_name:
            title = text.name_from_uri(file_name)

        meta = code_meta(title, extra, **options)
        result.update(meta)
        del result["code"]

        iscc_code_obj = compose(
            [meta["code"], content["code"], data["code"], instance["code"]]
        )
        result["iscc"] = iscc_code_obj.code
        concat = bytes.fromhex(result["metahash"]) + bytes.fromhex(result["datahash"])
        result["tophash"] = blake3(concat).hexdigest()
    finally:
        file_obj.close()

    return result


def code_meta(title, extra=None, **options):
    # type: (Union[str, bytes], Optional[Union[str, bytes, dict]], **Any) -> dict
    """Generate Meta Code from title and extra metadata.

    :param title: Used as input for first half of Meta-Code (or full if no extra)
    :param extra: Extended (str or dict) metadata used as input for second half of code.
    :key meta_bits: Length of generated Meta-Code in bits
    :key meta_ngram_size: Number of characters for sliding window over metadata text.
    :key meta_trim_title: Trim title length to this mumber of bytes
    :key meta_trim_extra: Trim extra data to this number of bytes
    :returns: Dict keys: code, title, matahash, (extra)
    """
    opts = Options(**options)
    nbits = opts.meta_bits
    nbytes = nbits // 8
    title_norm = text.normalize_text(title)

    if isinstance(extra, (str, bytes, bytearray)):
        try:
            extra = json.loads(extra)
        except JSONDecodeError:
            pass

    if extra is None:
        extra = ""
    elif isinstance(extra, dict):
        if "@context" in extra:
            extra = jsonld.normalize(
                extra, {"algorithm": "URDNA2015", "format": "application/n-quads"}
            )
        else:
            extra = jcs.canonicalize(extra)

    extra_norm = text.normalize_text(extra)
    title_trimmed = text.trim_text(title_norm, opts.meta_trim_title)
    extra_trimmed = text.trim_text(extra_norm, opts.meta_trim_extra)
    meta_simhash = meta_hash(title_trimmed, extra_trimmed, **options)
    header = write_header(MT.META, ST.NONE, VS.V0, nbits)
    digest = header + meta_simhash[:nbytes]
    code = encode_base32(digest)
    result = dict(
        code=code,
        title=title_trimmed,
    )
    if extra_trimmed:
        result["extra"] = extra_trimmed
    payload = title_trimmed + extra_trimmed
    result["metahash"] = blake3(payload.encode("utf-8")).hexdigest()
    return result


def code_content(data, **options):
    # type: (Union[Uri, File], **Any) -> dict
    """Detect mediatype and create corresponding Content-Code."""
    mediatype = mime_guess(data)
    gmt = mime_to_gmt(mediatype)

    if gmt == GMT.text:
        cc = code_text(data, **options)
    elif gmt == GMT.image:
        cc = code_image(data, **options)
    elif gmt == GMT.audio:
        cc = code_audio(data, **options)
    elif gmt == GMT.video:
        cc = code_video(data, **options)
    else:
        raise ValueError("Unknown mediatype")

    cc["mediatype"] = mediatype
    cc["gmt"] = gmt

    return cc


def code_text(data, **options):
    # type: (Readable, **Any) -> dict
    """Generate Content-ID Text

    :param data: Any kind of text document
    :key text_guess_title: whether to guess the title from the text itself as fallback.
    :key meta_trim_title: Max number of bytes for utf-8 encoded title.
    :key text_avg_chunk_size: Avg. number of chars per text chunk to be hashed.
    :key text_ngram_size: Sliding window size in number of characters.
    :key text_granular: Wether to extract granular text features
    """
    opts = Options(**options)
    nbits = opts.text_bits
    nbytes = nbits // 8
    result = {}

    f = uread.open_data(data)
    metadata = text.extract_text_metadata(f, **options)
    result.update(metadata)
    text_raw = text.extract_text(f)
    text_norm = text.normalize_text(text_raw)
    text_hash = text.hash_text(text_norm, **options)

    header = write_header(MT.CONTENT, ST_CC.TEXT, VS.V0, nbits)
    code = encode_base32(header + text_hash[:nbytes])
    result["code"] = code

    if opts.text_granular or opts.all_granular:
        features = text.extract_text_features(text_norm, **options)
        result["features"] = features

    return result


def code_image(data, **options):
    # type: (Union[Uri, Data, Image.Image], **Any) -> dict

    opts = Options(**options)
    nbits = opts.image_bits
    nbytes = nbits // 8
    assert nbits in (32, 64), "Content-ID Image does not yet support more than 64-bits"

    try:
        result = image.extract_image_metadata(data) or {}
    except Exception as e:
        logger.error(f"Failed image metadata extraction: {e}")
        result = {}

    if isinstance(data, Image.Image):
        img_obj = data
    else:
        with uread.open_data(data) as infile:
            img_obj = Image.open(io.BytesIO(infile.read()))

    if opts.image_exif_transpose:
        img_obj = exif_transpose(img_obj)

    width, height = img_obj.size
    result.update(dict(width=width, height=height))

    if opts.image_trim:
        img_obj = image.trim_image(img_obj)
        if img_obj.size != result.values():
            tw, th = img_obj.size
            result["trimmed"] = dict(width=tw, height=th)

    if opts.image_preview or opts.all_preview:
        preview = image.extract_image_preview(img_obj, **options)
        preview_uri = image.encode_image_to_data_uri(preview, **options)
        result["preview"] = preview_uri

    pixels = image.normalize_image(img_obj)
    hash_digest = image.hash_image(pixels)[:nbytes]
    header = write_header(MT.CONTENT, ST_CC.IMAGE, VS.V0, nbits)
    code = encode_base32(header + hash_digest)
    result["code"] = code

    return result


def code_audio(data, **options):
    # type: (Union[Uri, Data, List], **Any) -> dict
    """Generate Audio-ID from file(path) or Chromaprint features"""
    opts = Options(**options)
    result = dict()
    nbits = opts.audio_bits
    nbytes = nbits // 8
    if isinstance(data, list):
        chroma = dict(fingerprint=data)
    else:
        chroma = audio.extract_audio_features(data, **options)
        result.update(video.extract_video_metadata(data))

    shash_digest = audio.hash_audio(chroma["fingerprint"])

    if opts.audio_granular or opts.all_granular:
        features = audio.encode_audio_features(chroma["fingerprint"])
        result["features"] = features

    header = write_header(MT.CONTENT, ST_CC.AUDIO, VS.V0, nbits)
    code = encode_base32(header + shash_digest[:nbytes])
    result["code"] = code
    return result


def code_video(uri, **options):
    # type: (Union[Uri, File], **Any) -> dict
    """Compute Content-ID video."""
    opts = Options(**options)
    nbits = opts.video_bits

    result = {}
    metadata = video.extract_video_metadata(uri)
    result.update(metadata)

    crop_value = video.detect_video_crop(uri) if opts.video_crop else None
    signature = video.extract_video_signature(uri, crop_value, **options)
    logger.debug(f"mp7 signature size {naturalsize(len(signature))}")
    frames = read_ffmpeg_signature(signature)
    logger.debug(f"mp7 signature frames {len(frames)}")
    features = [tuple(sig.vector.tolist()) for sig in frames]
    video_hash = video.hash_video(features, **options)
    video_code = Code((MT.CONTENT, ST_CC.VIDEO, VS.V0, nbits, video_hash))
    result["code"] = video_code.code

    # Not yet part of schema
    # result["signature_fps"] = opts.video_fps
    # if crop_value:
    #     result["crop"] = crop_value.lstrip("crop=")

    if opts.video_include_fingerprint:
        result["fingerprint"] = base64.b64encode(signature).decode("ascii")

    if opts.video_preview or opts.all_preview:
        img_raw = video.extract_video_preview(uri)
        result["preview"] = image.encode_image_to_data_uri(img_raw)

    if opts.video_granular is False and opts.all_granular is False:
        return result

    if opts.video_scenes:
        cutpoints = video.detect_video_scenes(uri, **options)
        features = video.compute_video_features_scenes(frames, cutpoints)
    else:
        features = video.compute_video_features_rolling(frames, **options)

    result["features"] = features

    return result


def code_data(data, **options):
    # type: (Readable, **Any) -> dict
    """Create ISCC Data-Code.

    The Data-Code is a similarity preserving hash of the input data.

    :param data: File, filepath or raw data used for Data-Code creation.
    :key data_avg_chunk_size: Target chunk size in bytes for data chunking.
    :key data_granular: Return granular features (one hash per chunk).
    :key data_granular_factor: Size of granular data chunks times average chunk size.
    :key io_chunk_size: Number of bytes to read per IO operation.
    :return:
    """

    opts = Options(**options)
    nbits = opts.data_bits
    nbytes = nbits // 8
    features = []
    sizes = []

    for chunk in data_chunks(data, **options):
        sizes.append(len(chunk))
        features.append(xxhash.xxh32_intdigest(chunk))

    data_hash = minhash_256(features)
    header = write_header(MT.DATA, ST.NONE, VS.V0, nbits)
    code = encode_base32(header + data_hash[:nbytes])
    result = dict(code=code)

    if opts.data_granular or opts.all_granular:
        sizes, features = extract_data_features(data, **options)
        features = encode_data_features(sizes, features)
        result["features"] = features

    return result


def code_instance(data, **options):
    # type: (Readable, **Any) -> dict
    """Create ISCC Instance-Code.

    The Instance-Code is prefix of a cryptographic hash (blake3) of the input data.

    :param data: File, filepath or raw data used for Instance-Code creation.
    :key instance_bits: Length of generated Instance-Code in bits (default 64).
    :key io_chunk_size: Number of bytes to read per IO operation.
    :return: An InstanceCode object with attributes: code, datahash, filesize
    """
    opts = Options(**options)
    nbits = opts.instance_bits
    nbytes = nbits // 8
    filesize = 0
    b3 = blake3()
    stream = uread.open_data(data)

    buffer = stream.read(opts.io_chunk_size)
    while buffer:
        filesize += len(buffer)
        b3.update(buffer)
        buffer = stream.read(opts.io_chunk_size)

    datahash_digest = b3.digest()
    header = write_header(MT.INSTANCE, ST.NONE, VS.V0, nbits)
    code = encode_base32(header + datahash_digest[:nbytes])
    datahash = datahash_digest.hex()

    return dict(code=code, datahash=datahash, filesize=filesize)



import cython
from libc.stdint cimport uint8_t, uint32_t

cdef uint32_t AVG_SIZE_DATA
cdef uint32_t READ_SIZE
cdef uint32_t[256] GEAR

@cython.locals(pattern=uint32_t, i=uint32_t, size=uint32_t, barrier=uint32_t)
cdef uint32_t cdc_offset(const uint8_t[:], uint32_t mi, uint32_t ma, uint32_t cs, uint32_t mask_s, uint32_t mask_l)



# -*- coding: utf-8 -*-
import io
from typing import Union
from iscc.schema import Readable, Uri, File, Data
from typing import BinaryIO


def open_data(data):
    # type: (Readable) -> Union[BinaryIO]
    """Open filepath, rawdata or file-like object."""
    if isinstance(data, Uri.__args__):
        return open(str(data), "rb")
    elif isinstance(data, Data.__args__):
        return io.BytesIO(data)
    elif isinstance(data, File.__args__):
        data.seek(0)
        return data
    else:
        raise ValueError(f"Unupported data type {type(data)}")



# -*- coding: utf-8 -*-
"""Persistent inverted index for ISCCs, components and granular features.

Subdatabases
1.  iscc_none_v0_256
----------------------------
2.  meta_code_v0_64
3.  content_text_v0_64
4.  content_image_v0_64
4.  content audio_v0_64
5.  content video_v0_64
6.  data_code_v0_64
8.  instance_code_v0_64
----------------------------
9.  feat_text_v0_64
10. feat_image_v0_64
11. feat_audio_v0_64
12. feat_video_v0_64
----------------------------
13. metdata
"""
from os.path import join
from typing import List, Optional, Dict
from loguru import logger as log
import iscc
import lmdb
import struct
import shutil
from humanize import naturalsize


pack = lambda n: struct.pack("I", n)
unpack = lambda b: struct.unpack("I", b)[0]


class Index:
    def __init__(self, name="iscc-db", **options):
        """Create or open existing index.

        :param str name: Name of index.
        :param bool index_components: Create inverted index of components -> iscc
        :param bool index_features: Create inverted index of features -> iscc
        :param bool store_metadata: Store metadata in index.
        """

        self.name = name
        self.dbpath = join(iscc.APP_DIR, self.name)
        log.info(f"init storage at {self.dbpath}")
        self.env = lmdb.open(
            path=self.dbpath,
            map_size=2 ** 20,
            max_dbs=24,
            metasync=False,
            sync=False,
            readahead=False,
            writemap=False,
            meminit=False,
            map_async=True,
        )

    @property
    def map_size(self) -> int:
        return self.env.info()["map_size"]

    def add(self, code, features=None):
        # type: (str, Optional[List[Dict]]) -> int
        """Add an ISCC to the index."""

        # Check for duplicate
        exists = self.get_id(code)
        if exists is not None:
            return exists

        # Normalize
        components = iscc.decompose(code)
        iscc_obj = iscc.compose(components)

        # Add full code sequence to main index
        db = self.db_isccs()
        mainkey = self.next_key()
        self.put(db, mainkey, iscc_obj.bytes)

        # Add components to components index
        db = self.db_components()
        items = [(code.bytes, mainkey) for code in components]
        self.putmulti(db, items)

        # Add feature hashes
        features = [] if features is None else features
        for fobj in features:
            db = self.db_features(kind=fobj["kind"])
            items = [(iscc.decode_base64(f), mainkey) for f in fobj["features"]]
            self.putmulti(db, items)

        return unpack(mainkey)

    def get_id(self, code) -> Optional[int]:
        """Get internal ID of for ISCC"""
        # Find per component matches
        components = iscc.decompose(code)
        db = self.db_components()
        idxs = []
        with self.env.begin(db=db) as txn:
            for code in components:
                idx = txn.get(code.bytes)
                if idx is not None:
                    idxs.append(idx)
        # Check if any of the full code entries is an exact match
        full_code_bytes = iscc.compose(components).bytes
        db = self.db_isccs()
        with self.env.begin(db=db) as txn:
            for idx in idxs:
                if txn.get(idx) == full_code_bytes:
                    return unpack(idx)

    def add_component(self, code, pk):
        # type: (iscc.Code, bytes) -> None
        db = self.db(code.type_id.encode("ascii"))
        self.put(db, code.hash_bytes, pk)

    def next_key(self) -> bytes:
        """Next free autoincrement id as 4-byte key"""
        db = self.db_isccs()
        with self.env.begin(db) as txn:
            with txn.cursor(db) as c:
                empty = not c.last()
                idx = unpack(c.key()) + 1 if not empty else 0
        return pack(idx)

    def put(self, db, key: bytes, value: bytes):
        try:
            with self.env.begin(db, write=True) as txn:
                txn.put(key, value)
        except lmdb.MapFullError:
            new_size = self.map_size * 2
            log.info(f"Resizing {self.dbpath} to {naturalsize(new_size)}")
            self.env.set_mapsize(self.map_size * 2)
            with self.env.begin(db, write=True) as txn:
                txn.put(key, value)

    def putmulti(self, db, items, dupdata=True, overwrite=True):
        try:
            with self.env.begin(db, write=True) as txn:
                with txn.cursor(db) as c:
                    c.putmulti(items, dupdata=dupdata, overwrite=overwrite)
        except lmdb.MapFullError:
            new_size = self.map_size * 2
            log.info(f"Resizing {self.dbpath} to {naturalsize(new_size)}")
            self.env.set_mapsize(self.map_size * 2)
            with self.env.begin(db, write=True) as txn:
                with txn.cursor(db) as c:
                    c.putmulti(items, dupdata=dupdata, overwrite=overwrite)

    def isccs(self):
        """Iterates over indexed ISCC codes in insertion order."""
        db = self.db_isccs()
        with self.env.begin(db) as txn:
            with txn.cursor(db) as c:
                while c.next():
                    yield c.value()

    def iscc(self, idx: int) -> bytes:
        """Get ISCC by index id"""
        db = self.db_isccs()
        with self.env.begin(db) as txn:
            return txn.get(pack(idx))

    def components(self):
        """Itereates over indexed components."""
        db = self.db_components()
        with self.env.begin(db) as txn:
            with txn.cursor(db) as c:
                for key in c.iternext_nodup():
                    yield key

    def dbs(self) -> List[str]:
        """Return a list of existing sub-databases in the index."""
        dbnames = []
        with self.env.begin() as txn:
            c = txn.cursor()
            while c.next():
                dbnames.append(c.key())
        return dbnames

    def destory(self):
        """Close and delete index from disk."""
        self.env.close()
        shutil.rmtree(self.dbpath)

    def close(self):
        """Close index."""
        self.env.close()

    def db_isccs(self):
        return self.env.open_db(b"isccs", integerkey=True, create=True)

    def db_components(self):
        return self.env.open_db(
            b"components", dupsort=True, integerdup=True, create=True, dupfixed=True
        )

    def db_features(self, kind):
        return self.env.open_db(
            kind.encode("utf-8"),
            dupsort=True,
            integerkey=True,
            integerdup=True,
            create=True,
            dupfixed=True,
        )

    def __len__(self):
        """Number of indexed ISCCs"""
        db = self.db_isccs()
        with self.env.begin(db) as txn:
            stat = txn.stat(db)
            return stat.get("entries")

    def __contains__(self, item):
        """Check if full iscc code is in index."""
        id_ = self.get_id(item)
        return False if id_ is None else True


if __name__ == "__main__":
    idx = Index()
    len(idx)



# -*- coding: utf-8 -*-
import os
from pathlib import Path
import click

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata

__version__ = importlib_metadata.version(__name__)

HERE = Path(__file__).parent
APP_NAME = "iscc"
APP_DIR = click.get_app_dir(APP_NAME, roaming=False)
os.makedirs(APP_DIR, exist_ok=True)
os.environ["TIKA_STARTUP_MAX_RETRY"] = "20"
os.environ["LOGURU_AUTOINIT"] = "False"

from tika import tika

tika.log.disabled = True

from iscc.core import *
from iscc.metrics import *
from iscc.image import *
from iscc.minhash import minhash
from iscc.text import *
from iscc.wtahash import *
from iscc.data import *
from iscc.codec import *
from iscc.video import *
from iscc.jcs import *
from iscc.match import *
from iscc.mediatype import *
from iscc.index import *



# -*- coding: utf-8 -*-
"""Functions that measure haming distance of compact binary codes"""
from typing import Union
from bitarray import bitarray
from bitarray.util import count_xor, hex2ba
from iscc.codec import Code, read_header, decode_base32, decode_base64, decompose, MT


__all__ = ["distance", "compare"]

# ISCC code in various possible representations
Icode = Union[str, bytes, int, Code]


def distance(a, b, mixed=False):
    # type: (Icode, Icode, bool) -> int
    """Calculate hamming distance for singular ISCC component codes of same type.

    I mixed is true, we allow for comparison of hashes with different length and
    subtypes. If components have different length we compute the hamming distance of
    the shorter code with the prefix of the longer code.

    ISCC Codes can be supplied in these formats:
    - Code: a codec.Code object
    - str: base32 encoded ISCC component(including header).
    - bytes: raw byte digest of ISCC component (without header).
    - int: integer representation of ISCC component (without header).
    """
    if isinstance(a, Code):
        return distance_code(a.code, b.code, mixed=mixed)
    if isinstance(a, str):
        return distance_code(a, b, mixed=mixed)
    if isinstance(a, bytes):
        if mixed:
            nbytes = min((len(a), len(b)))
            a, b = a[:nbytes], b[:nbytes]
        return distance_bytes(a, b)
    if isinstance(a, int):
        return distance_int(a, b)


def distance_code(a, b, mixed=False):
    # type: (str, str, bool) -> int
    """
    Calculate hamming distance for ISCC component codes.

    Component type, subtype, version and lenght must be the same if mixed is 'False'.
    If mixed is 'True' we allow for different subtypes and length. In case codes are of
    different length and mixed is 'true' we calculate common length prefix distance.
    """
    a, b = decode_base32(a), decode_base32(b)
    a, b = read_header(a), read_header(b)
    if mixed:
        assert a[0] == b[0] and a[2] == b[2], "Maintype and version do not match."
        # calculate common prefix length distance
        nbytes = min((a[3], b[3])) // 8
        return distance_bytes(a[-1][:nbytes], b[-1][:nbytes])
    else:
        assert a[:-1] == b[:-1], "Code header values do not match."
        # calculate strict distance (ignore subsequent codes)
        a = a[-1][: a[3] // 8]
        b = b[-1][: b[3] // 8]
        return distance_bytes(a, b)


def distance_int(a, b):
    # type: (int, int) -> int
    """Calculate hamming dinstance for integer hashes."""
    return bin(a ^ b).count("1")


def distance_bytes(a, b):
    # type: (bytes, bytes) -> int
    """Calculate hamming distance for bytes hash digests of equal length."""
    return count_xor(hex2ba(a.hex()), hex2ba(b.hex()))


def distance_hex(a, b):
    # type: (str, str) -> int
    """Calculate hamming distnace for hex encoded hashes of equal length."""
    return count_xor(hex2ba(a), hex2ba(b))


def distance_ba(a, b):
    # type: (bitarray, bitarray) -> int
    """Calculate hamming distance for bitarray objects of equal length."""
    return count_xor(a, b)


def distance_b64(a, b):
    # type: (str, str) -> int
    """Calculate hamming distance for base64 encoded integers"""
    return distance_bytes(decode_base64(a), decode_base64(b))


def compare(a, b):
    # type: (str, str) -> dict
    """Calculate separate hamming distances of compatible components of two ISCCs

    :returns: A dict with mdist, cdist, ddist, imatch keys
    """
    ac = decompose(a)
    bc = decompose(b)
    result = {}
    for ca in ac:
        for cb in bc:
            cat = (ca.maintype, ca.subtype, ca.version)
            cbt = (cb.maintype, cb.subtype, ca.version)
            if cat == cbt:
                if ca.maintype != MT.INSTANCE:
                    result[ca.type_id[0].lower() + "dist"] = distance(
                        ca, cb, mixed=True
                    )
                else:
                    result["imatch"] = ca.hash_bytes == cb.hash_bytes
    return result



# -*- coding: utf-8 -*-
import requests
from loguru import logger
import hashlib
import io
import os
import re
from typing import Sequence, Generator
from urllib.parse import urlparse
from iscc import APP_DIR


def sliding_window(seq, width):
    # type: (Sequence, int) -> Generator[Sequence[Sequence]]
    """
    Generate a sequence of equal "width" slices each advancing by one elemnt.
    All types that have a length and can be sliced are supported (list, tuple, str ...).
    The result type matches the type of the input sequence.
    Fragment slices smaller than the width at the end of the sequence are not produced.
    If "witdh" is smaller than the input sequence than one element will be returned that
    is shorter than the requested width.
    """
    assert width >= 2, "Sliding window width must be 2 or bigger."
    idx = range(max(len(seq) - width + 1, 1))
    return (seq[i : i + width] for i in idx)


def download_file(url, md5=None, sanitize=False):
    """Download file to app dir and return path."""
    url_obj = urlparse(url)
    file_name = os.path.basename(url_obj.path)
    if sanitize:
        file_name = safe_filename(file_name)
    out_path = os.path.join(APP_DIR, file_name)
    if os.path.exists(out_path):
        logger.debug(f"Already downloaded: {file_name}")
        if md5:
            md5_calc = hashlib.md5(open(out_path, "rb").read()).hexdigest()
            assert md5 == md5_calc, f"Integrity error for {out_path}"
            return out_path

    r = requests.get(url, stream=True)
    chunk_size = 1024
    iter_size = 0
    with io.open(out_path, "wb") as fd:
        for chunk in r.iter_content(chunk_size):
            fd.write(chunk)
            iter_size += chunk_size

    if md5:
        md5_calc = hashlib.md5(open(out_path, "rb").read()).hexdigest()
        assert md5 == md5_calc, f"Integrity error for {out_path}"
    return out_path


def safe_filename(s: str, max_len: int = 255) -> str:
    """Sanitize a string making it safe to use as a filename.
    See: https://en.wikipedia.org/wiki/Filename.
    """
    ntfs_chars = [chr(i) for i in range(0, 31)]
    chars = [
        r'"',
        r"\#",
        r"\$",
        r"\%",
        r"'",
        r"\*",
        r"\,",
        r"\.",
        r"\/",
        r"\:",
        r'"',
        r"\;",
        r"\<",
        r"\>",
        r"\?",
        r"\\",
        r"\^",
        r"\|",
        r"\~",
        r"\\\\",
    ]
    pattern = "|".join(ntfs_chars + chars)
    regex = re.compile(pattern, re.UNICODE)
    fname = regex.sub("", s)
    return fname[:max_len].rsplit(" ", 0)[0]



# -*- coding: utf-8 -*-
from json import JSONDecodeError
from loguru import logger
import json
import os
import shutil
import subprocess
import platform
import tarfile
import zipfile
import stat
from more_itertools import chunked, windowed
import iscc
from iscc.schema import FeatureType, Options, Readable
from iscc.simhash import similarity_hash
from iscc.utils import download_file
from iscc.codec import encode_base64
from typing import Any, List
from iscc import uread


def hash_audio(features):
    # type: (List[int]) -> bytes
    """Create 256-bit audio similarity hash from chromaprint vector"""
    digests = []
    for int_features in windowed(features, 8, fillvalue=0):
        digest = b""
        for int_feature in int_features:
            digest += int_feature.to_bytes(4, "big", signed=True)
        digests.append(digest)
    shash_digest = similarity_hash(digests)
    return shash_digest


def extract_audio_features(data, **options):
    # type: (Readable, **Any) -> dict
    """Returns Chromaprint fingerprint.

    A dictionary with keys:
    - duration: total duration of extracted fingerprint in seconds (Broken with pipe)
    - fingerprint: 32-bit (4 byte) integers as features
    """
    opts = Options(**options)
    length = str(opts.audio_max_duration)
    file = uread.open_data(data)
    cmd = [fpcalc_bin(), "-raw", "-json", "-signed", "-length", length, "-"]
    res = subprocess.run(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, input=file.read()
    )
    output = res.stdout.decode("utf-8")
    try:
        result = json.loads(output)
    except JSONDecodeError:
        raise RuntimeError(f'Fpcalc error: {res.stderr.decode("utf-8")}')

    return result


def encode_audio_features(features):
    # type: (List[int]) -> dict
    """Pack into 64-bit base encoded features.

    Note: Last feature will be 32-bit if number of features is uneven.
    """
    fingerprints = []
    for int_features in chunked(features, 2):
        digest = b""
        for int_feature in int_features:
            digest += int_feature.to_bytes(4, "big", signed=True)
        fingerprints.append(encode_base64(digest))
    return dict(kind=FeatureType.audio.value, features=fingerprints)


FPCALC_VERSION = "1.5.0"
FPCALC_URL_BASE = (
    f"https://github.com/acoustid/chromaprint/releases/download/v{FPCALC_VERSION}/"
)

FPCALC_OS_MAP = {
    "Linux": "chromaprint-fpcalc-{}-linux-x86_64.tar.gz".format(FPCALC_VERSION),
    "Darwin": "chromaprint-fpcalc-{}-macos-x86_64.tar.gz".format(FPCALC_VERSION),
    "Windows": "chromaprint-fpcalc-{}-windows-x86_64.zip".format(FPCALC_VERSION),
}


def fpcalc_bin():
    """Returns local path to fpcalc executable."""
    if platform.system() == "Windows":
        return os.path.join(iscc.APP_DIR, "fpcalc-{}.exe".format(FPCALC_VERSION))
    return os.path.join(iscc.APP_DIR, "fpcalc-{}".format(FPCALC_VERSION))


def fpcalc_is_installed():
    """"Check if fpcalc is installed."""
    fp = fpcalc_bin()
    return os.path.isfile(fp) and os.access(fp, os.X_OK)


def fpcalc_download_url():
    """Return system and version dependant download url"""
    return os.path.join(FPCALC_URL_BASE, FPCALC_OS_MAP[platform.system()])


def fpcalc_download():
    """Download fpcalc and return path to archive file."""
    return download_file(fpcalc_download_url())


def fpcalc_extract(archive):
    """Extract archive with fpcalc executable."""
    if archive.endswith(".zip"):
        with zipfile.ZipFile(archive, "r") as zip_file:
            for member in zip_file.namelist():
                filename = os.path.basename(member)
                if filename == "fpcalc.exe":
                    source = zip_file.open(member)
                    target = open(fpcalc_bin(), "wb")
                    with source, target:
                        shutil.copyfileobj(source, target)
    elif archive.endswith("tar.gz"):
        with tarfile.open(archive, "r:gz") as tar_file:
            for member in tar_file.getmembers():
                if member.isfile() and member.name.endswith("fpcalc"):
                    source = tar_file.extractfile(member)
                    target = open(fpcalc_bin(), "wb")
                    with source, target:
                        shutil.copyfileobj(source, target)


def fpcalc_install():
    """Install fpcalc command line tool and return path to executable."""
    if fpcalc_is_installed():
        logger.debug("Fpcalc is already installed.")
        return fpcalc_bin()
    archive_path = fpcalc_download()
    fpcalc_extract(archive_path)
    st = os.stat(fpcalc_bin())
    os.chmod(fpcalc_bin(), st.st_mode | stat.S_IEXEC)
    assert fpcalc_is_installed()
    return fpcalc_bin()


def fpcalc_version_info():
    """Get fpcalc version"""
    try:
        r = subprocess.run([fpcalc_bin(), "-v"], stdout=subprocess.PIPE)
        return r.stdout.decode("utf-8").strip().split()[2]
    except FileNotFoundError:
        return "FPCALC not installed"



# -*- coding: utf-8 -*-
import json
import sys
from loguru import logger as log
import typer
import iscc
from pathlib import Path
from typing import Optional
from codetiming import Timer


app = typer.Typer()

Parg = typer.Argument(..., help="Path to file or folder (default: currend dir)")
Debug = typer.Option(False, "--debug", "-d", help="Show debug output.")
Verbose = typer.Option(0, "--verbose", "-v", count=True)


@app.command()
def explain(icode: str):
    """Decode ISCC to human readable format."""
    code = iscc.Code(icode)
    typer.echo(code.explain)


@app.command()
def distance(a: str, b: str):
    """Calculate hamming distance of codes"""
    codes_a = iscc.decompose(a)
    codes_b = iscc.decompose(b)
    for a, b in zip(codes_a, codes_b):
        typer.echo(
            f"Distance {a.maintype.name}: {a.code} - {b.code} = {iscc.distance(a, b)} bits"
        )


@app.command()
def decompose(icode: str):
    """Decompose ISCC into its components"""
    codes = iscc.decompose(icode)
    typer.echo(f'ISCC:{"-".join(c.code for c in codes)}')


@app.command("iscc")
def main(path: Path, verbose: int = Verbose, debug: bool = Debug):
    """Generate ISCC for file(s)."""
    if debug:
        log.add(sys.stderr)
        log.info("Debug messages activated!")
    for file in files(path):
        try:
            with Timer(text="ISCC generation time: {:0.4f} seconds", logger=log.info):
                result = iscc.code_iscc(file)
            if not isinstance(result, dict):
                typer.echo(f"no result for {file.name} ({result})")
                continue
            if verbose == 0:
                typer.echo(f"ISCC:{result['iscc']}")
            elif verbose == 1:
                typer.echo(f"ISCC:{result['iscc']},{result['filename']}")
            elif verbose == 2:
                typer.echo(result)
            else:
                typer.echo(json.dumps(result, indent=2))
        except Exception as e:
            log.error(f"{file.name} failed with: {str(e).strip()}")


def files(path: Optional[Path]):
    """Build a list of files from a path"""

    if path is None:
        return list(f for f in Path.cwd().glob("*") if f.is_file())
    elif path.is_file():
        return [path]
    elif path.is_dir():
        return list(f for f in path.glob("*") if f.is_file())
    else:
        typer.Abort("Nothing to do.")


if __name__ == "__main__":
    app()



# -*- coding: utf-8 -*-
"""Fuction for similarity preserving Data-Code."""
from typing import Any, Tuple, List
import xxhash
from more_itertools import chunked
from iscc.cdc import data_chunks
from iscc.minhash import minhash_64, minhash_256
from iscc.codec import encode_base64
from iscc.schema import FeatureType, Readable, Options


def hash_data(data):
    _, features = extract_data_features(data)
    return hash_data_features(features)


def hash_data_features(features):
    # type: (List[int]) -> bytes
    """Create 256-bit data similarity hash from data-chunk int32 hashes"""
    return minhash_256(features)


def extract_data_features(data, **options):
    # type: (Readable, **Any) -> Tuple[List[int], List[int]]
    """Extract low level data features (chunk-sizes, int32-hashes)"""
    opts = Options(**options)
    features = []
    sizes = []
    for chunk in data_chunks(data, **opts.dict()):
        sizes.append(len(chunk))
        features.append(xxhash.xxh32_intdigest(chunk))
    return sizes, features


def encode_data_features(sizes, features, **options):
    # type: (List[int], List[int], **Any) -> dict
    """Reduce and encode low level features.

    :return: dicttionary with {"sizes": ..., "features": ...}
    """
    opts = Options(**options)
    encoded_sizes = [sum(fh) for fh in chunked(sizes, opts.data_granular_factor)]
    encoded_features = [
        encode_base64(minhash_64(cf))
        for cf in chunked(features, opts.data_granular_factor)
    ]
    return dict(
        kind=FeatureType.data,
        features=encoded_features,
        sizes=encoded_sizes,
    )



# -*- coding: utf-8 -*-
from loguru import logger
from typing import List, Optional, Union
import mimetypes
import magic
from PIL import Image
from iscc.schema import GMT, Readable
from iscc import uread


__all__ = [
    "mime_guess",
    "mime_normalize",
    "mime_supported",
    "mime_clean",
    "mime_to_gmt",
    "mime_from_name",
    "mime_from_data",
]


def mime_guess(data, file_name=None):
    # type: (Readable, str) -> str
    """Heuristic guessing of mediatype for different kinds of inputs.

    We try matching by file extension. If that fails we match by content sniffing.
    """

    guess_name, guess_data = None, None
    file = uread.open_data(data)

    if file_name is None:
        if hasattr(file, "name"):
            file_name = file.name
        elif hasattr(file, "filename"):
            file_name = file.filename

    if file_name:
        guess_name = mime_from_name(file_name)

    guess_data = mime_from_data(file.read(4096))

    # Normalize
    guess_data = mime_normalize(guess_data)
    guess_name = mime_normalize(guess_name)

    return guess_name or guess_data


def mime_normalize(mime: str) -> str:
    """Return normalized version of a mediatype."""
    return MEDIATYPE_NORM.get(mime, mime)


def mime_supported(mime: str) -> bool:
    """Check if mediatype is supported"""
    return mime_normalize(mime) in SUPPORTED_MEDIATYPES


def mime_from_name(name: str) -> Optional[str]:
    """Guess mediatype from filename or url."""
    return mimetypes.guess_type(name)[0]


def mime_from_data(data: bytes) -> Optional[str]:
    """Guess mediatype by sniffing raw header data."""
    return magic.from_buffer(data, mime=True)


def mime_clean(mime: Union[str, List]):
    """
    Clean mimetype/content-type string or first entry of a list of mimetype strings.
    Also removes semicolon separated encoding information.
    """
    if mime and isinstance(mime, List):
        mime = mime[0]
    if mime:
        mime = mime.split(";")[0]
    return mime.strip()


def mime_to_gmt(mime_type: str, file_path=None):
    """Get generic mediatype from mimetype."""
    mime_type = mime_clean(mime_type)
    if mime_type == "image/gif" and file_path:
        img = Image.open(file_path)
        if img.is_animated:
            return "video"
        else:
            return "image"
    entry = SUPPORTED_MEDIATYPES.get(mime_type)
    if entry:
        return entry["gmt"]
    gmt = mime_type.split("/")[0]
    if gmt in list(GMT):
        logger.warning(f"Guessing GMT from {mime_type}")
        return gmt


mimetypes.add_type("text/markdown", ".md")
mimetypes.add_type("text/markdown", ".markdown")
mimetypes.add_type("application/x-mobipocket-ebook", ".mobi")
mimetypes.add_type("application/x-sqlite3", ".sqlite")


SUPPORTED_MEDIATYPES = {
    # Text Formats
    "application/rtf": {"gmt": "text", "ext": "rtf"},
    "application/msword": {"gmt": "text", "ext": "doc"},
    "application/pdf": {"gmt": "text", "ext": "pdf"},
    "application/epub+zip": {"gmt": "text", "ext": "epub"},
    "text/xml": {"gmt": "text", "ext": "xml"},
    "application/json": {"gmt": "text", "ext": "json"},
    "application/xhtml+xml": {"gmt": "text", "ext": "xhtml"},
    "application/vnd.oasis.opendocument.text": {"gmt": "text", "ext": "odt"},
    "text/html": {"gmt": "text", "ext": "html"},
    "text/plain": {"gmt": "text", "ext": "txt"},
    "application/x-ibooks+zip": {"gmt": "text", "ext": "ibooks"},
    "text/markdown": {"gmt": "text", "ext": ["md", "markdown"]},
    "application/vnd.openxmlformats-officedocument.wordprocessingml.document": {
        "gmt": "text",
        "ext": "docx",
    },
    "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet": {
        "gmt": "text",
        "ext": "xlsx",
    },
    # Note: pptx only detected by file extension. Sniffing gives 'application/zip'
    "application/vnd.openxmlformats-officedocument.presentationml.presentation": {
        "gmt": "text",
        "ext": "pptx",
    },
    "application/vnd.ms-excel": {"gmt": "text", "ext": "xls"},
    # TODO add mobi support
    # "application/x-mobipocket-ebook": {
    #     "gmt": "text",
    #     "ext": ["mobi", "prc", "azw", "azw3", "azw4"],
    # },
    # Image Formats
    "image/bmp": {"gmt": "image", "ext": "bmp"},
    "image/gif": {"gmt": "image", "ext": "gif"},
    "image/jpeg": {"gmt": "image", "ext": ["jpg", "jpeg"]},
    "image/png": {"gmt": "image", "ext": "png"},
    "image/tiff": {"gmt": "image", "ext": "tif"},
    "image/vnd.adobe.photoshop": {"gmt": "image", "ext": "psd"},
    "application/postscript": {"gmt": "image", "ext": "eps"},
    # Audio Formats
    "audio/mpeg": {"gmt": "audio", "ext": "mp3"},
    "audio/wav": {"gmt": "audio", "ext": "wav"},
    "audio/ogg": {"gmt": "audio", "ext": "ogg"},
    "audio/aiff": {"gmt": "audio", "ext": "aif"},
    "audio/x-aiff": {"gmt": "audio", "ext": "aif"},
    "audio/x-flac": {"gmt": "audio", "ext": "flac"},
    "audio/opus": {"gmt": "audio", "ext": "opus"},
    # Video Formats
    "application/vnd.rn-realmedia": {"gmt": "video", "ext": "rm"},
    "video/x-dirac": {"gmt": "video", "ext": "drc"},
    "video/3gpp": {"gmt": "video", "ext": "3gp"},
    "video/3gpp2": {"gmt": "video", "ext": "3g2"},
    "video/x-ms-asf": {"gmt": "video", "ext": "asf"},
    "video/avi": {"gmt": "video", "ext": "avi"},
    "video/webm": {"gmt": "video", "ext": "webm"},
    "video/mpeg": {"gmt": "video", "ext": ["mpeg", "mpg", "m1v", "vob"]},
    "video/mp4": {"gmt": "video", "ext": "mp4"},
    "video/x-m4v": {"gmt": "video", "ext": "m4v"},
    "video/x-matroska": {"gmt": "video", "ext": "mkv"},
    "video/ogg": {"gmt": "video", "ext": ["ogg", "ogv"]},
    "video/quicktime": {"gmt": "video", "ext": ["mov", "f4v"]},
    "video/x-flv": {"gmt": "video", "ext": "flv"},
    "application/x-shockwave-flash": {"gmt": "video", "ext": "swf"},
    "video/h264": {"gmt": "video", "ext": "h264"},
    "video/x-ms-wmv": {"gmt": "video", "ext": "wmv"},
}

MEDIATYPE_NORM = {
    "audio/x-aiff": "audio/aiff",
    "image/x-ms-bmp": "image/bmp",
    "video/x-msvideo": "video/avi",
}

SUPPORTED_EXTENSIONS = []
for v in SUPPORTED_MEDIATYPES.values():
    ext = v["ext"]
    if isinstance(ext, str):
        SUPPORTED_EXTENSIONS.append(ext)
    else:
        for e in ext:
            SUPPORTED_EXTENSIONS.append(e)
